package com.sun.xml.internal.bind.v2.schemagen;

import com.sun.xml.internal.bind.v2.schemagen.xmlschema.Particle;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.Occurs;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.TypeDefParticle;
import com.sun.xml.internal.bind.v2.schemagen.xmlschema.ContentModelContainer;
import java.util.Iterator;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.List;

abstract class Tree
{
    Tree() {
    }
    
    Tree makeOptional(boolean really) {
        return really ? new Tree.Optional(this) : this;
    }
    
    Tree makeRepeated(boolean really) {
        return really ? new Tree.Repeated(this) : this;
    }
    
    static Tree makeGroup(GroupKind kind, List<Tree> children) {
        if (children.size() == 1) {
            return (Tree)children.get(0);
        }
        List<Tree> normalizedChildren = new ArrayList(children.size());
        for (Tree t : children) {
            if ((t instanceof Tree.Group)) {
                Tree.Group g = (Tree.Group)t;
                if (g.kind == kind) {
                    normalizedChildren.addAll(Arrays.asList(g.children));
                    continue;
                }
            }
            normalizedChildren.add(t);
        }
        return new Tree.Group(kind, (Tree[])normalizedChildren.toArray(new Tree[normalizedChildren.size()]));
    }
    
    abstract boolean isNullable();
    
    boolean canBeTopLevel() {
        return false;
    }
    
    protected abstract void write(ContentModelContainer p0, boolean p1, boolean p2);
    
    protected void write(TypeDefParticle ct) {
        if (this.canBeTopLevel()) {
            this.write((ContentModelContainer)ct._cast(ContentModelContainer.class), false, false);
        }
        else {
            new Tree.Group(GroupKind.SEQUENCE, new Tree[] { this }).write(ct);
        }
    }
    
    protected final void writeOccurs(Occurs o, boolean isOptional, boolean repeated) {
        if (isOptional) {
            o.minOccurs(0);
        }
        if (repeated) {
            o.maxOccurs("unbounded");
        }
    }
    
    abstract static class Term extends Tree
    {
        @Override
        boolean isNullable() {
            return false;
        }
    }
    
    private static final class Optional extends Tree
    {
        private final Tree body;
        
        private Optional(Tree body) {
            this.body = body;
        }
        
        @Override
        boolean isNullable() {
            return true;
        }
        
        @Override
        Tree makeOptional(boolean really) {
            return this;
        }
        
        @Override
        protected void write(ContentModelContainer parent, boolean isOptional, boolean repeated) {
            this.body.write(parent, true, repeated);
        }
    }
    
    private static final class Repeated extends Tree
    {
        private final Tree body;
        
        private Repeated(Tree body) {
            this.body = body;
        }
        
        @Override
        boolean isNullable() {
            return this.body.isNullable();
        }
        
        @Override
        Tree makeRepeated(boolean really) {
            return this;
        }
        
        @Override
        protected void write(ContentModelContainer parent, boolean isOptional, boolean repeated) {
            this.body.write(parent, isOptional, true);
        }
    }
    
    private static final class Group extends Tree
    {
        private final GroupKind kind;
        private final Tree[] children;
        
        private Group(GroupKind kind, Tree... children) {
            this.kind = kind;
            this.children = children;
        }
        
        @Override
        boolean canBeTopLevel() {
            return true;
        }
        
        @Override
        boolean isNullable() {
            if (this.kind == GroupKind.CHOICE) {
                Tree[] children2 = this.children;
                int length2 = children2.length;
                int j = 0;
                while (j < length2) {
                    Tree t = children2[j];
                    if (t.isNullable()) {
                        return true;
                    }
                    ++j;
                }
                return false;
            }
            Tree[] children = this.children;
            int length = children.length;
            int i = 0;
            while (i < length) {
                Tree t = children[i];
                if (!t.isNullable()) {
                    return false;
                }
                ++i;
            }
            return true;
        }
        
        @Override
        protected void write(ContentModelContainer parent, boolean isOptional, boolean repeated) {
            Particle c = this.kind.write(parent);
            this.writeOccurs(c, isOptional, repeated);
            for (Tree child : this.children) {
                child.write(c, false, false);
            }
        }
    }
}
